/**
 * Introduction
 *
 * @title ROV - Easy applications from C
 *
 * @copyright Jason Tribbeck / 7th ARM 1994-2001
 *
 * @authors Jason Tribbeck
 *
 * @file preference.c
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "main.h"
#include "canvas.h"
#include "icons.h"
#include "draw.h"
#include "msg.h"

#include "preference.h"

#define TXTH (56)
#define TXTW (48)

#define PREF_BOT (96)
#define PREF_BW (PREFERENCE_MAXX/4)
#define PREF_BX(xx) ((20+PREF_BW*xx+17)&0xfffffffe)

static canvas _pb, * _pf;
static icon _save, _restore, _cancel, _set;
static char * * _t;
static int * _y, * _x, * _w;
static int _maxpage;
static _rov_prefh * _p, _ph;
static int _pageopened = 0, _po = 0, _r = 1;
static int preference_bodge = 2;

static void preference_reopen(int page, int flag);

static void preference_pbredraw(canvas c, int x0, int y0, int x1, int y1)
{
  int i;
  icon_data d;

  d.indirecttext.valid = (char *) - 1;
  d.indirecttext.length = 12;
  draw_gcol(0);
  draw_move(18, PREF_BOT - 2);
  draw_lineto(0, PREFERENCE_MAXY + 4);
  /* draw_lineto(PREFERENCE_MAXX + 4, 0); */
  draw_lineto(_x[_pageopened] + 4, 0);
  draw_moveto(_w[_pageopened] - 6 + preference_bodge, 0);
  draw_line(PREFERENCE_MAXX + 20, PREFERENCE_MAXY + PREF_BOT + 2);
  draw_move(16, PREF_BOT - 4);
  draw_lineto(0, PREFERENCE_MAXY + 8);
  /* draw_lineto(PREFERENCE_MAXX + 8, 0); */
  draw_lineto(_x[_pageopened] + 6, 0);
  draw_moveto(_w[_pageopened] - 6 + preference_bodge, 0);
  draw_line(PREFERENCE_MAXX + 22, PREFERENCE_MAXY + PREF_BOT + 4);
  draw_gcol(4);
  draw_move(18, PREF_BOT - 2);
  draw_lineto(PREFERENCE_MAXX + 4, 0);
  draw_lineto(0, PREFERENCE_MAXY + 4);
  draw_move(16, PREF_BOT - 4);
  draw_lineto(PREFERENCE_MAXX + 8, 0);
  draw_lineto(0, PREFERENCE_MAXY + 8);
  draw_gcol(0);
  for(i = 0; i < _maxpage; i ++) {
    if(i != _pageopened) {
      draw_gcol(0);
      draw_move(_x[i] + 20, _y[i] * TXTH + PREFERENCE_MAXY + PREF_BOT + 4);
      draw_lineto(0, - TXTH + 2);
      draw_moveto(2, 0);
      draw_lineto(0, TXTH - 4);
      draw_lineto(_w[i] - 6 + preference_bodge, 0);
      draw_moveto(0, 2);
      draw_lineto(- _w[i] + 6 - preference_bodge, 0);
      draw_moveto(_w[i] - 6 + preference_bodge, - 2);
      draw_gcol(4);
      draw_lineto(0, - TXTH + 4);
      draw_moveto(2, 0);
      draw_lineto(0, TXTH - 2);
      d.indirecttext.text = _t[i];
      draw_ploticon(_x[i] + 20, PREFERENCE_MAXY + PREF_BOT + 2 + _y[i] * TXTH - TXTH, _x[i] + _w[i] + 20, PREFERENCE_MAXY + PREF_BOT + 2 + _y[i] * TXTH, 0x13000119, & d);
    }   else {
      draw_gcol(0);
      draw_move(_x[_pageopened] + 20, _y[_pageopened] * TXTH + PREFERENCE_MAXY + PREF_BOT + 8);
      draw_lineto(0, - TXTH - 2);
      draw_moveto(2, 0);
      draw_lineto(0, TXTH);
      draw_lineto(_w[_pageopened] - 6 + preference_bodge, 0);
      draw_moveto(0, 2);
      draw_lineto(- _w[_pageopened] + 6 - preference_bodge, 0);
      draw_moveto(_w[_pageopened] - 6 + preference_bodge, - 2);
      draw_gcol(4);
      draw_lineto(0, - TXTH - 2);
      draw_moveto(2, 2);
      draw_lineto(0, TXTH + 2);
      d.indirecttext.text = _t[_pageopened];
      /* draw_gcol(1);
	 draw_move(_x[_pageopened] + 24, PREFERENCE_MAXY + PREF_BOT + 4);
	 draw_lineto(_w[_pageopened] - 10 + preference_bodge, 0);
	 draw_moveto(0, - 2);
	 draw_lineto(- _w[_pageopened] + 10 - preference_bodge, 0); */
      draw_ploticon(_x[_pageopened] + 20, PREFERENCE_MAXY + PREF_BOT + 6 + _y[_pageopened] * TXTH - TXTH, _x[_pageopened] + _w[_pageopened] + 20, PREFERENCE_MAXY + PREF_BOT + 6 + _y[_pageopened] * TXTH, 0x17000119, & d);
    }
  }
}

void preference_close(void)
{
  _p[_pageopened](PREFERENCE_CANCEL, _pageopened);
  canvas_close(_pf[_pageopened]);
  canvas_close(_pb);
  _po = 0;
}

static void preference_wopen(canvas c, canvas_openstr * o)
{
  canvas_openstr o2;
  memcpy(& o2, o, sizeof(canvas_openstr));
  o2.x0 += 20;
  o2.x1 -= 20;
  o2.y0 += PREF_BOT;
  o2.y1 = o2.y0 + PREFERENCE_MAXY;
  o2.scy = 0;
  o2.scx = 0;
  canvas_open(_pf[_pageopened], & o2);
  o -> behind = canvas_handle(_pf[_pageopened]);
  canvas_open(_pb, o);
}

static void preference_click(canvas c, canvas_mousestr * m)
{
  int y, i;

  y = (m -> y - PREF_BOT - PREFERENCE_MAXY) / TXTH + 1;
  if((y > 0) && (y <= _r) && (m -> x > 20) && (m -> x < 20 + PREFERENCE_MAXX)) {
    for(i = 0; i < _maxpage; i ++) {
      if((_y[i] == y) && (m -> x >= _x[i] + 20) && (m -> x <= _x[i] + 20 + _w[i])) {
	preference_reopen(i, 0);
	return;
      }
    }
  }
}

void preference_save(canvas c, canvas_mousestr * m)
{
  _ph(PREFERENCE_SAVE, _pageopened);
}

void preference_cancel(void)
{
  _ph(PREFERENCE_CANCEL, _pageopened);
  canvas_close(_pb);
  canvas_close(_pf[_pageopened]);
}

void preference_restore(void)
{
  _ph(PREFERENCE_DEFAULT, _pageopened);
}

void preference_set(canvas c, icon i, canvas_mousestr * m)
{
  if(m -> b == 4) {
    canvas_close(_pb);
    canvas_close(_pf[_pageopened]);
    _ph(PREFERENCE_SETANDCLOSE, _pageopened);
  } else {
    _ph(PREFERENCE_SET, _pageopened);
  }
}

canvas preference_init(int pages, char * title, _rov_prefh han)
{
  int i;

  _maxpage = pages;
  _t = rcalloc(sizeof(char *), pages);
  _x = rcalloc(sizeof(int), pages);
  _y = rcalloc(sizeof(int), pages);
  _w = rcalloc(sizeof(int), pages);
  _p = rcalloc(sizeof(_rov_prefh), pages);
  _pf = rcalloc(sizeof(canvas), pages);
  for(i = 0; i < pages; i ++) {
    _t[i] = 0;
    _y[i] = -1;
    _x[i] = 0;
    _w[i] = 0;
    _p[i] = 0;
  }

  _pb = canvas_create(PREFERENCE_MAXX + 40, PREFERENCE_MAXY + 600, title);
  canvas_addhandler(_pb, CANVAS_OPENHANDLER, preference_wopen);
  canvas_addhandler(_pb, CANVAS_CLOSEHANDLER, preference_close);
  canvas_addhandler(_pb, CANVAS_REDRAWHANDLER, preference_pbredraw);
  canvas_addhandler(_pb, CANVAS_CLICKHANDLER, preference_click);
  canvas_make(_pb);

  _restore = icon_createbutton(_pb, PREF_BX(0), 16, msg_lookuprom("Default:Default"), ((PREF_BW - 32) + 1) & 0xfffffffe);
  _save = icon_createbutton(_pb, PREF_BX(1), 16, msg_lookuprom("Save:Save"), ((PREF_BW - 32) + 1) & 0xfffffffe);
  _cancel = icon_createbutton(_pb, PREF_BX(2), 16, msg_lookuprom("Cancel:Cancel"), ((PREF_BW - 32) + 1) & 0xfffffffe);
  _set = icon_createdefaultbutton(_pb, PREF_BX(3), 16, msg_lookuprom("Set:Set"), ((PREF_BW - 32) + 1) & 0xfffffffe);

  icon_addhandler(_save, ICON_CLICKHANDLER, preference_save);
  icon_addhandler(_restore, ICON_CLICKHANDLER, preference_restore);
  icon_addhandler(_cancel, ICON_CLICKHANDLER, preference_cancel);
  icon_addhandler(_set, ICON_CLICKHANDLER, preference_set);
  _ph = han;

  return _pb;
}

static void preference_swappages(int src, int dest)
{
  int i;

  for(i = 0; i < _maxpage; i ++) {
    if(_y[i] == src) {
      _y[i] = dest;
    } else {
      if(_y[i] == dest) _y[i] = src;
    }
  }
}

static void preference_reopen(int page, int flag)
{
  canvas_openstr o;
  int behind = 0;

  if((page == _pageopened) && (!flag)) return;

  if(page == PREFERENCE_NULLPAGE) {
    page = _pageopened;
  }
  if(page != _pageopened) {
    if(!flag) {
      canvas_getcanvasstate(_pf[_pageopened], & o);
      behind = o.behind;
      canvas_close(_pf[_pageopened]);
    }
    if(_y[page] != 1) {
      preference_swappages(_y[page], _y[_pageopened]);
      canvas_redraw(_pb, 20, PREFERENCE_MAXY + PREF_BOT, PREFERENCE_MAXX + 20, PREFERENCE_MAXY + PREF_BOT + _r * TXTH + TXTH + 8);
    } else {
      canvas_redraw(_pb, 20 + _x[page], PREFERENCE_MAXY + PREF_BOT, 20 + preference_bodge + _x[page] + _w[page], PREFERENCE_MAXY + PREF_BOT + TXTH + 12);
      canvas_redraw(_pb, 20 + _x[_pageopened], PREFERENCE_MAXY + PREF_BOT, 20 + preference_bodge + _x[_pageopened] + _w[_pageopened], PREFERENCE_MAXY + PREF_BOT + TXTH + 12);
    }
    if(!flag) {
      _p[_pageopened](PREFERENCE_CLOSE, _pageopened);
      _p[_pageopened](PREFERENCE_CHANGE, page);
    }
  }
  if(flag) _p[page](PREFERENCE_NEW, page);
  _pageopened = page;
  if(_po) {
    if(flag) canvas_reopen(_pb);
  } else {
    canvas_getcanvasstate(_pb, & o);
    o.scx = o.scy = 0;
    o.x0 = (draw_maxx - PREFERENCE_MAXX - 40) >> 1;
    o.x1 = (draw_maxx + PREFERENCE_MAXX + 40) >> 1;
    o.y0 = (draw_maxy - PREFERENCE_MAXY - _r * TXTH - 20 - PREF_BOT) >> 1;
    o.y1 = (draw_maxy + PREFERENCE_MAXY + _r * TXTH + 40 + PREF_BOT) >> 1;
    if(flag) o.behind = -1; else o.behind = behind;
    canvas_open(_pb, & o);
    _po = 1;
  }
  if(!flag) {
    canvas_getcanvasstate(_pb, & o);
    o.behind = behind;
  } else {
    canvas_getcanvasstate(_pb, & o);
  }
  o.x0 += 20;
  o.x1 -= 20;
  o.y0 += PREF_BOT;
  o.y1 = o.y0 + PREFERENCE_MAXY;
  o.scx = 0;
  o.scy = 0;
  canvas_open(_pf[_pageopened], & o);
  canvas_claimcaret(_pb);
}

static void preference_realignrow(int row)
{
  int start = -1, end = 0, i, x, to_give;

  for(i = 0; i < _maxpage; i ++) {
    if((_y[i] == row) && (_x[i] == 0)) start = i;
    if(_y[i] == row) end = i;
  }

  if(start != -1) {
    to_give = PREFERENCE_MAXX + 2; /* Bodge factor ! */
    x = 0;
    for(i = start; i <= end; i ++) {
      to_give -= _w[i] + 2;
    }
    for(i = start; i <= end; i ++) {
      _x[i] = x;
      if((_w[i] + start) & 1) start -= 1;
      start = to_give / (end + 1 - i);
      _w[i] += start;
      to_give -= start;
      x += _w[i] + 2;
    }
  }
}

canvas preference_makepage(char * str, _rov_prefh p)
{
  int i;

  for(i = 0; i < _maxpage; i ++) {
    if(_y[i] == -1) {
      _t[i] = rcalloc(1, main_strlen(str) + 1);
      main_strcpy(_t[i], str);
      _w[i] = (main_strwidth(str) + TXTW + 1) & 0xfffffffe;
      if(_x[i] + _w[i] > PREFERENCE_MAXX) {
	preference_realignrow(_r ++);
	_x[i] = 0;
      }
      _y[i] = _r;
      if(i != _maxpage - 1) {
	_x[i + 1] = _x[i] + _w[i];
      } else {
	if(_r != 1) preference_realignrow(_r); else preference_bodge = 0;
      }
      _p[i] = p;
      _pf[i] = canvas_create(PREFERENCE_MAXX, PREFERENCE_MAXY, "");
      canvas_setflags(_pf[i], 0xff000030U, 0x80000020U);
      canvas_setoutlinecolour(_pf[i], 255);
      canvas_make(_pf[i]);
      return _pf[i];
    }
  }
  return 0;
}

canvas preference_makecanvaspage(char * str, _rov_prefh p, canvas c)
{
  int i;

  for(i = 0; i < _maxpage; i ++) {
    if(_y[i] == -1) {
      _t[i] = rcalloc(1, main_strlen(str) + 1);
      main_strcpy(_t[i], str);
      _w[i] = (main_strwidth(str) + TXTW + 1) & 0xfffffffe;
      if(_x[i] + _w[i] > PREFERENCE_MAXX) {
	preference_realignrow(_r ++);
	_x[i] = 0;
      }
      _y[i] = _r;
      if(i != _maxpage - 1) {
	_x[i + 1] = _x[i] + _w[i];
      } else {
	if(_r != 1) preference_realignrow(_r); else preference_bodge = 0;
      }
      _p[i] = p;
      _pf[i] = c;
      return _pf[i];
    }
  }
  return 0;
}

void preference_open(int page)
{
  canvas_close(_pf[_pageopened]);
  preference_reopen(page, 1);
}
